@@ -1,52 +1,51 @@
 from abc import (ABC,
                  abstractmethod)
 from itertools import groupby
-from numbers import Rational
 from operator import attrgetter
-from typing import (List,
+from typing import (Iterable,
+                    List,
                     Optional,
+                    Sequence,
+                    Tuple,
                     Union as Union_)
 
+from ground.base import Context
+from ground.hints import (Point,
+                          Segment)
 from reprit.base import generate_repr
 
-from clipping.hints import (HolelessMix,
-                            Multipoint,
-                            Multiregion,
-                            Multisegment)
-from . import bounding_box
+from . import bounding
 from .event import (ShapedEvent as Event,
+                    event_to_segment_endpoints,
                     events_to_connectivity)
-from .events_queue import (BinaryEventsQueueKey as EventsQueueKey,
-                           HolelessEventsQueue as EventsQueue)
+from .events_queue import HolelessEventsQueue as EventsQueue
+from .hints import (HolelessMix,
+                    Multiregion,
+                    SegmentEndpoints)
 from .sweep_line import BinarySweepLine as SweepLine
 from .utils import (all_equal,
-                    contour_to_oriented_segments,
+                    contour_to_oriented_edges_endpoints,
+                    endpoints_to_segments,
                     pairwise,
-                    to_multiregion_base,
-                    to_multiregion_x_max,
-                    to_rational_multiregion)
+                    to_multiregion_x_max)
 
 
 class Operation(ABC):
-    __slots__ = 'left', 'right', 'accurate', '_events_queue'
+    __slots__ = 'context', 'left', 'right', '_events_queue'
 
     def __init__(self,
                  left: Multiregion,
                  right: Multiregion,
-                 accurate: bool) -> None:
+                 context: Context) -> None:
         """
         Initializes operation.
 
         :param left: left operand.
         :param right: right operand.
-        :param accurate:
-            flag that tells whether to use slow but more accurate arithmetic
-            for floating point numbers.
+        :param context: operation context.
         """
-        self.left = left
-        self.right = right
-        self.accurate = accurate
-        self._events_queue = EventsQueue()
+        self.context, self.left, self.right = context, left, right
+        self._events_queue = EventsQueue(context)
 
     __repr__ = generate_repr(__init__)
 
@@ -65,27 +64,60 @@ class Operation(ABC):
                                             else below_event.interior_to_left)
         event.in_result = self.in_result(event)
 
+    def events_to_multiregion(self, events: Iterable[Event]) -> Multiregion:
+        events = sorted([event for event in events if event.primary.in_result],
+                        key=self._events_queue.key)
+        for index, event in enumerate(events):
+            event.position = index
+        processed = [False] * len(events)
+        result = []
+        connectivity = events_to_connectivity(events)
+        contour_cls = self.context.contour_cls
+        for index, event in enumerate(events):
+            if processed[index]:
+                continue
+            contour_start = event.start
+            vertices = [contour_start]
+            contour_events = [event]
+            cursor = event
+            complement_position = event.complement.position
+            processed[index] = processed[complement_position] = True
+            while cursor.end != contour_start:
+                vertices.append(cursor.end)
+                position = _to_next_position(complement_position, processed,
+                                             connectivity)
+                if position is None:
+                    break
+                cursor = events[position]
+                contour_events.append(cursor)
+                complement_position = cursor.complement.position
+                processed[position] = processed[complement_position] = True
+            result.append(contour_cls(vertices))
+        return result
+
     def fill_queue(self) -> None:
         events_queue = self._events_queue
         for region in self.left:
-            events_queue.register_segments(
-                    contour_to_oriented_segments(region), True)
+            events_queue.register(
+                    contour_to_oriented_edges_endpoints(region,
+                                                        context=self.context),
+                    True)
         for region in self.right:
-            events_queue.register_segments(
-                    contour_to_oriented_segments(region), False)
+            events_queue.register(
+                    contour_to_oriented_edges_endpoints(region,
+                                                        context=self.context),
+                    False)
 
     @abstractmethod
     def in_result(self, event: Event) -> bool:
         """Detects if event will be presented in result of the operation."""
 
     def normalize_operands(self) -> None:
-        left, right = self.left, self.right
-        if self.accurate and not issubclass(to_multiregion_base(left + right),
-                                            Rational):
-            self.left, self.right = (to_rational_multiregion(left),
-                                     to_rational_multiregion(right))
+        pass
 
-    def process_event(self, event: Event, processed_events: List[Event],
+    def process_event(self,
+                      event: Event,
+                      processed_events: List[Event],
                       sweep_line: SweepLine) -> None:
         if event.is_right_endpoint:
             processed_events.append(event)
@@ -120,23 +152,48 @@ class CompleteIntersection(Operation):
     __slots__ = ()
 
     def compute(self) -> HolelessMix:
+        points, segments, multiregion = self._compute()
+        return (self.context.multipoint_cls(points),
+                self.context.multisegment_cls(segments), multiregion)
+
+    def in_result(self, event: Event) -> bool:
+        return (event.inside
+                or not event.from_left and event.is_common_region_boundary)
+
+    def sweep(self) -> Iterable[Event]:
+        self.fill_queue()
+        result = []
+        sweep_line = SweepLine(self.context)
+        min_max_x = min(to_multiregion_x_max(self.left),
+                        to_multiregion_x_max(self.right))
+        while self._events_queue:
+            event = self._events_queue.pop()
+            if min_max_x < event.start.x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
+
+    def _compute(self) -> Tuple[Sequence[Point], Sequence[Segment],
+                                Multiregion]:
         if not (self.left and self.right):
             return [], [], []
-        left_bounding_box = bounding_box.from_multiregion(self.left)
-        right_bounding_box = bounding_box.from_multiregion(self.right)
-        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
+        left_box = bounding.from_multiregion(self.left,
+                                             context=self.context)
+        right_box = bounding.from_multiregion(self.right,
+                                              context=self.context)
+        if bounding.disjoint_with(left_box, right_box):
             return [], [], []
-        self.left = bounding_box.to_intersecting_regions(right_bounding_box,
-                                                         self.left)
-        self.right = bounding_box.to_intersecting_regions(left_bounding_box,
-                                                          self.right)
+        self.left = bounding.to_intersecting_regions(right_box, self.left,
+                                                     context=self.context)
+        self.right = bounding.to_intersecting_regions(left_box, self.right,
+                                                      context=self.context)
         if not (self.left and self.right):
             return [], [], []
         self.normalize_operands()
         events = sorted(self.sweep(),
-                        key=EventsQueueKey)
-        multipoint = []  # type: Multipoint
-        multisegment = []  # type: Multisegment
+                        key=self._events_queue.key)
+        points = []  # type: List[Point]
+        endpoints = []  # type: List[SegmentEndpoints]
         for start, same_start_events in groupby(events,
                                                 key=attrgetter('start')):
             same_start_events = list(same_start_events)
@@ -147,34 +204,20 @@ class CompleteIntersection(Operation):
                 no_segment_found = True
                 for event, next_event in pairwise(same_start_events):
                     if (event.from_left is not next_event.from_left
-                            and event.segment == next_event.segment):
+                            and event.start == next_event.start
+                            and event.end == next_event.end):
                         no_segment_found = False
                         if not event.is_right_endpoint:
-                            multisegment.append(next_event.segment)
+                            endpoints.append(
+                                    event_to_segment_endpoints(next_event))
                 if no_segment_found and all(not event.complement.in_result
                                             if event.is_right_endpoint
                                             else not event.in_result
                                             for event in same_start_events):
-                    multipoint.append(start)
-        return multipoint, multisegment, events_to_multiregion(events)
-
-    def sweep(self) -> List[Event]:
-        self.fill_queue()
-        result = []
-        sweep_line = SweepLine()
-        min_max_x = min(to_multiregion_x_max(self.left),
-                        to_multiregion_x_max(self.right))
-        while self._events_queue:
-            event = self._events_queue.pop()
-            start_x, _ = event.start
-            if start_x > min_max_x:
-                break
-            self.process_event(event, result, sweep_line)
-        return result
-
-    def in_result(self, event: Event) -> bool:
-        return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
+                    points.append(start)
+        return (points, endpoints_to_segments(endpoints,
+                                              context=self.context),
+                self.events_to_multiregion(events))
 
 
 class Intersection(Operation):
@@ -183,29 +226,30 @@ class Intersection(Operation):
     def compute(self) -> Multiregion:
         if not (self.left and self.right):
             return []
-        left_bounding_box = bounding_box.from_multiregion(self.left)
-        right_bounding_box = bounding_box.from_multiregion(self.right)
-        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
+        left_box = bounding.from_multiregion(self.left,
+                                             context=self.context)
+        right_box = bounding.from_multiregion(self.right,
+                                              context=self.context)
+        if bounding.disjoint_with(left_box, right_box):
             return []
-        self.left = bounding_box.to_coupled_regions(right_bounding_box,
-                                                    self.left)
-        self.right = bounding_box.to_coupled_regions(left_bounding_box,
-                                                     self.right)
+        self.left = bounding.to_coupled_regions(right_box, self.left,
+                                                context=self.context)
+        self.right = bounding.to_coupled_regions(left_box, self.right,
+                                                 context=self.context)
         if not (self.left and self.right):
             return []
         self.normalize_operands()
-        return events_to_multiregion(self.sweep())
+        return self.events_to_multiregion(self.sweep())
 
-    def sweep(self) -> List[Event]:
+    def sweep(self) -> Iterable[Event]:
         self.fill_queue()
         result = []
-        sweep_line = SweepLine()
+        sweep_line = SweepLine(self.context)
         min_max_x = min(to_multiregion_x_max(self.left),
                         to_multiregion_x_max(self.right))
         while self._events_queue:
             event = self._events_queue.pop()
-            start_x, _ = event.start
-            if start_x > min_max_x:
+            if min_max_x < event.start.x:
                 break
             self.process_event(event, result, sweep_line)
         return result
@@ -215,40 +259,9 @@ class Intersection(Operation):
                 or not event.from_left and event.is_common_region_boundary)
 
 
-def events_to_multiregion(events: List[Event]) -> Multiregion:
-    events = sorted([event for event in events if event.primary.in_result],
-                    key=EventsQueueKey)
-    for index, event in enumerate(events):
-        event.position = index
-    processed = [False] * len(events)
-    result = []
-    connectivity = events_to_connectivity(events)
-    for index, event in enumerate(events):
-        if processed[index]:
-            continue
-        contour_start = event.start
-        contour = [contour_start]
-        contour_events = [event]
-        cursor = event
-        complement_position = event.complement.position
-        processed[index] = processed[complement_position] = True
-        while cursor.end != contour_start:
-            contour.append(cursor.end)
-            position = _to_next_position(complement_position, processed,
-                                         connectivity)
-            if position is None:
-                break
-            cursor = events[position]
-            contour_events.append(cursor)
-            complement_position = cursor.complement.position
-            processed[position] = processed[complement_position] = True
-        result.append(contour)
-    return result
-
-
 def _to_next_position(position: int,
                       processed: List[bool],
-                      connectivity: List[int]) -> Optional[int]:
+                      connectivity: Sequence[int]) -> Optional[int]:
     candidate = position
     while True:
         candidate = connectivity[candidate]
