@@ -1,60 +1,59 @@
 from abc import (ABC,
                  abstractmethod)
 from itertools import groupby
-from numbers import Rational
 from operator import attrgetter
-from typing import (List,
+from typing import (Iterable,
+                    List,
                     Optional,
+                    Sequence,
+                    Tuple,
                     Union as Union_)
 
+from ground.base import Context
+from ground.hints import (Multipolygon,
+                          Point,
+                          Polygon,
+                          Segment)
 from reprit.base import generate_repr
 
-from clipping.hints import (Contour,
-                            Mix,
-                            Multipoint,
-                            Multipolygon,
-                            Multisegment)
-from . import bounding_box
+from . import bounding
 from .event import (HoleyEvent as Event,
+                    event_to_segment_endpoints,
                     events_to_connectivity)
-from .events_queue import (BinaryEventsQueueKey as EventsQueueKey,
-                           HoleyEventsQueue as EventsQueue)
+from .events_queue import HoleyEventsQueue as EventsQueue
+from .hints import (Mix,
+                    SegmentEndpoints)
 from .sweep_line import BinarySweepLine as SweepLine
 from .utils import (all_equal,
+                    endpoints_to_segments,
                     pairwise,
-                    polygon_to_oriented_segments,
+                    polygon_to_oriented_edges_endpoints,
                     shrink_collinear_vertices,
-                    to_first_boundary_vertex,
-                    to_multipolygon_base,
-                    to_multipolygon_x_max,
-                    to_rational_multipolygon)
+                    to_first_border_vertex,
+                    to_polygons_x_max)
 
 
 class Operation(ABC):
-    __slots__ = 'left', 'right', 'accurate', '_events_queue'
+    __slots__ = 'context', 'left', 'right', '_events_queue'
 
     def __init__(self,
-                 left: Multipolygon,
-                 right: Multipolygon,
-                 accurate: bool) -> None:
+                 left: Sequence[Polygon],
+                 right: Sequence[Polygon],
+                 context: Context) -> None:
         """
         Initializes operation.
 
         :param left: left operand.
         :param right: right operand.
-        :param accurate:
-            flag that tells whether to use slow but more accurate arithmetic
-            for floating point numbers.
+        :param context: operation context.
         """
-        self.left = left
-        self.right = right
-        self.accurate = accurate
-        self._events_queue = EventsQueue()
+        self.context, self.left, self.right = context, left, right
+        self._events_queue = EventsQueue(context)
 
     __repr__ = generate_repr(__init__)
 
     @abstractmethod
-    def compute(self) -> Union_[Multipolygon, Mix]:
+    def compute(self) -> Union_[Mix, Multipolygon]:
         """
         Computes result of the operation.
         """
@@ -72,28 +71,69 @@ class Operation(ABC):
                                            else below_event)
         event.in_result = self.in_result(event)
 
+    def events_to_polygons(self, events: Iterable[Event]) -> Sequence[Polygon]:
+        events = sorted([event for event in events if event.primary.in_result],
+                        key=self._events_queue.key)
+        for index, event in enumerate(events):
+            event.position = index
+        are_internal, depths, holes, parents = [], [], [], []
+        processed = [False] * len(events)
+        contour_cls = self.context.contour_cls
+        contours = []
+        connectivity = events_to_connectivity(events)
+        for index, event in enumerate(events):
+            if processed[index]:
+                continue
+            contour_id = len(contours)
+            _compute_relations(event, contour_id, are_internal, depths, holes,
+                               parents)
+            vertices = _events_to_contour_vertices(event, events, contour_id,
+                                                   connectivity, processed)
+            shrink_collinear_vertices(vertices,
+                                      context=self.context)
+            if depths[contour_id] % 2:
+                # holes will be in clockwise order
+                vertices.reverse()
+            contours.append(contour_cls(vertices))
+        result = []
+        polygon_cls = self.context.polygon_cls
+        for index, contour in enumerate(contours):
+            if are_internal[index]:
+                # hole of a hole is an external polygon
+                result.extend(
+                        polygon_cls(contours[hole_index],
+                                    [contours[hole_hole_index]
+                                     for hole_hole_index in holes[hole_index]])
+                        for hole_index in holes[index])
+            else:
+                result.append(polygon_cls(contour,
+                                          [contours[hole_index]
+                                           for hole_index in holes[index]]))
+        return result
+
     def fill_queue(self) -> None:
         events_queue = self._events_queue
         for polygon in self.left:
-            events_queue.register_segments(
-                    polygon_to_oriented_segments(polygon), True)
+            events_queue.register(
+                    polygon_to_oriented_edges_endpoints(polygon,
+                                                        context=self.context),
+                    True)
         for polygon in self.right:
-            events_queue.register_segments(
-                    polygon_to_oriented_segments(polygon), False)
+            events_queue.register(
+                    polygon_to_oriented_edges_endpoints(polygon,
+                                                        context=self.context),
+                    False)
 
     @abstractmethod
     def in_result(self, event: Event) -> bool:
         """Detects if event will be presented in result of the operation."""
 
     def normalize_operands(self) -> None:
-        left, right = self.left, self.right
-        if (self.accurate
-                and not issubclass(to_multipolygon_base(left + right),
-                                   Rational)):
-            self.left, self.right = (to_rational_multipolygon(left),
-                                     to_rational_multipolygon(right))
+        pass
 
-    def process_event(self, event: Event, processed_events: List[Event],
+    def process_event(self,
+                      event: Event,
+                      processed_events: List[Event],
                       sweep_line: SweepLine) -> None:
         if event.is_right_endpoint:
             processed_events.append(event)
@@ -123,10 +163,10 @@ class Operation(ABC):
                 self.compute_fields(below_event, below_below_event)
                 self.compute_fields(event, below_event)
 
-    def sweep(self) -> List[Event]:
+    def sweep(self) -> Iterable[Event]:
         self.fill_queue()
         result = []
-        sweep_line = SweepLine()
+        sweep_line = SweepLine(self.context)
         events_queue = self._events_queue
         while events_queue:
             self.process_event(events_queue.pop(), result, sweep_line)
@@ -137,60 +177,92 @@ class Difference(Operation):
     __slots__ = ()
 
     def compute(self) -> Multipolygon:
-        if not (self.left and self.right):
-            return self.left
-        left_bounding_box = bounding_box.from_multipolygon(self.left)
-        if bounding_box.disjoint_with(
-                left_bounding_box, bounding_box.from_multipolygon(self.right)):
-            return self.left
-        self.right = bounding_box.to_coupled_polygons(left_bounding_box,
-                                                      self.right)
-        if not self.right:
-            return self.left
-        self.normalize_operands()
-        return events_to_multipolygon(self.sweep())
+        return self.context.multipolygon_cls(self._compute())
 
     def in_result(self, event: Event) -> bool:
         return (event.outside
                 if event.from_left
                 else event.inside or event.is_common_polyline_component)
 
-    def sweep(self) -> List[Event]:
+    def sweep(self) -> Iterable[Event]:
         self.fill_queue()
         result = []
         events_queue = self._events_queue
-        sweep_line = SweepLine()
-        left_x_max = to_multipolygon_x_max(self.left)
+        sweep_line = SweepLine(self.context)
+        left_x_max = to_polygons_x_max(self.left)
         while events_queue:
             event = events_queue.pop()
-            start_x, _ = event.start
-            if start_x > left_x_max:
+            if left_x_max < event.start.x:
                 break
             self.process_event(event, result, sweep_line)
         return result
 
+    def _compute(self) -> Sequence[Polygon]:
+        if not (self.left and self.right):
+            return self.left
+        left_box = bounding.from_polygons(self.left,
+                                          context=self.context)
+        if bounding.disjoint_with(
+                left_box, bounding.from_polygons(self.right,
+                                                 context=self.context)):
+            return self.left
+        self.right = bounding.to_coupled_polygons(left_box, self.right,
+                                                  context=self.context)
+        if not self.right:
+            return self.left
+        self.normalize_operands()
+        return self.events_to_polygons(self.sweep())
+
 
 class CompleteIntersection(Operation):
     __slots__ = ()
 
     def compute(self) -> Mix:
+        points, segments, polygons = self._compute()
+        context = self.context
+        return (context.multipoint_cls(points),
+                context.multisegment_cls(segments),
+                context.multipolygon_cls(polygons))
+
+    def in_result(self, event: Event) -> bool:
+        return (event.inside
+                or not event.from_left and event.is_common_region_boundary)
+
+    def sweep(self) -> Iterable[Event]:
+        self.fill_queue()
+        result = []
+        events_queue = self._events_queue
+        sweep_line = SweepLine(self.context)
+        min_max_x = min(to_polygons_x_max(self.left),
+                        to_polygons_x_max(self.right))
+        while events_queue:
+            event = events_queue.pop()
+            if min_max_x < event.start.x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
+
+    def _compute(self) -> Tuple[Sequence[Point], Sequence[Segment],
+                                Sequence[Polygon]]:
         if not (self.left and self.right):
             return [], [], []
-        left_bounding_box = bounding_box.from_multipolygon(self.left)
-        right_bounding_box = bounding_box.from_multipolygon(self.right)
-        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
+        left_box = bounding.from_polygons(self.left,
+                                          context=self.context)
+        right_box = bounding.from_polygons(self.right,
+                                           context=self.context)
+        if bounding.disjoint_with(left_box, right_box):
             return [], [], []
-        self.left = bounding_box.to_intersecting_polygons(right_bounding_box,
-                                                          self.left)
-        self.right = bounding_box.to_intersecting_polygons(left_bounding_box,
-                                                           self.right)
+        self.left = bounding.to_intersecting_polygons(right_box, self.left,
+                                                      context=self.context)
+        self.right = bounding.to_intersecting_polygons(left_box, self.right,
+                                                       context=self.context)
         if not (self.left and self.right):
             return [], [], []
         self.normalize_operands()
         events = sorted(self.sweep(),
-                        key=EventsQueueKey)
-        multipoint = []  # type: Multipoint
-        multisegment = []  # type: Multisegment
+                        key=self._events_queue.key)
+        points = []  # type: List[Point]
+        endpoints = []  # type: List[SegmentEndpoints]
         for start, same_start_events in groupby(events,
                                                 key=attrgetter('start')):
             same_start_events = list(same_start_events)
@@ -201,149 +273,117 @@ class CompleteIntersection(Operation):
                 no_segment_found = True
                 for event, next_event in pairwise(same_start_events):
                     if (event.from_left is not next_event.from_left
-                            and event.segment == next_event.segment):
+                            and event.start == next_event.start
+                            and event.end == next_event.end):
                         no_segment_found = False
                         if not event.is_right_endpoint:
-                            multisegment.append(next_event.segment)
+                            endpoints.append(
+                                    event_to_segment_endpoints(next_event))
                 if no_segment_found and all(not event.primary.in_result
                                             for event in same_start_events):
-                    multipoint.append(start)
-        return multipoint, multisegment, events_to_multipolygon(events)
-
-    def sweep(self) -> List[Event]:
-        self.fill_queue()
-        result = []
-        events_queue = self._events_queue
-        sweep_line = SweepLine()
-        min_max_x = min(to_multipolygon_x_max(self.left),
-                        to_multipolygon_x_max(self.right))
-        while events_queue:
-            event = events_queue.pop()
-            start_x, _ = event.start
-            if start_x > min_max_x:
-                break
-            self.process_event(event, result, sweep_line)
-        return result
-
-    def in_result(self, event: Event) -> bool:
-        return (event.inside
-                or not event.from_left and event.is_common_region_boundary)
+                    points.append(start)
+        return (points,
+                endpoints_to_segments(endpoints,
+                                      context=self.context),
+                self.events_to_polygons(events))
 
 
 class Intersection(Operation):
     __slots__ = ()
 
     def compute(self) -> Multipolygon:
-        if not (self.left and self.right):
-            return []
-        left_bounding_box = bounding_box.from_multipolygon(self.left)
-        right_bounding_box = bounding_box.from_multipolygon(self.right)
-        if bounding_box.disjoint_with(left_bounding_box, right_bounding_box):
-            return []
-        self.left = bounding_box.to_coupled_polygons(right_bounding_box,
-                                                     self.left)
-        self.right = bounding_box.to_coupled_polygons(left_bounding_box,
-                                                      self.right)
-        if not (self.left and self.right):
-            return []
-        self.normalize_operands()
-        return events_to_multipolygon(self.sweep())
-
-    def sweep(self) -> List[Event]:
-        self.fill_queue()
-        result = []
-        events_queue = self._events_queue
-        sweep_line = SweepLine()
-        min_max_x = min(to_multipolygon_x_max(self.left),
-                        to_multipolygon_x_max(self.right))
-        while events_queue:
-            event = events_queue.pop()
-            start_x, _ = event.start
-            if start_x > min_max_x:
-                break
-            self.process_event(event, result, sweep_line)
-        return result
+        return self.context.multipolygon_cls(self._compute())
 
     def in_result(self, event: Event) -> bool:
         return (event.inside
                 or not event.from_left and event.is_common_region_boundary)
 
+    def sweep(self) -> Iterable[Event]:
+        self.fill_queue()
+        result = []
+        events_queue = self._events_queue
+        sweep_line = SweepLine(self.context)
+        min_max_x = min(to_polygons_x_max(self.left),
+                        to_polygons_x_max(self.right))
+        while events_queue:
+            event = events_queue.pop()
+            if min_max_x < event.start.x:
+                break
+            self.process_event(event, result, sweep_line)
+        return result
+
+    def _compute(self) -> Sequence[Polygon]:
+        if not (self.left and self.right):
+            return []
+        left_box = bounding.from_polygons(self.left,
+                                          context=self.context)
+        right_box = bounding.from_polygons(self.right,
+                                           context=self.context)
+        if bounding.disjoint_with(left_box, right_box):
+            return []
+        self.left = bounding.to_coupled_polygons(right_box, self.left,
+                                                 context=self.context)
+        self.right = bounding.to_coupled_polygons(left_box, self.right,
+                                                  context=self.context)
+        if not (self.left and self.right):
+            return []
+        self.normalize_operands()
+        return self.events_to_polygons(self.sweep())
+
 
 class SymmetricDifference(Operation):
     __slots__ = ()
 
     def compute(self) -> Multipolygon:
-        if not (self.left and self.right):
-            return self.left or self.right
-        elif bounding_box.disjoint_with(
-                bounding_box.from_multipolygon(self.left),
-                bounding_box.from_multipolygon(self.right)):
-            result = self.left + self.right
-            result.sort(key=to_first_boundary_vertex)
-            return result
-        self.normalize_operands()
-        return events_to_multipolygon(self.sweep())
+        return self.context.multipolygon_cls(self._compute())
 
     def in_result(self, event: Event) -> bool:
         return not event.is_overlap
 
+    def _compute(self) -> Sequence[Polygon]:
+        if not (self.left and self.right):
+            return self.left or self.right
+        elif bounding.disjoint_with(
+                bounding.from_polygons(self.left,
+                                       context=self.context),
+                bounding.from_polygons(self.right,
+                                       context=self.context)):
+            result = []
+            result += self.left
+            result += self.right
+            result.sort(key=to_first_border_vertex)
+            return result
+        self.normalize_operands()
+        return self.events_to_polygons(self.sweep())
+
 
 class Union(Operation):
     __slots__ = ()
 
     def compute(self) -> Multipolygon:
+        return self.context.multipolygon_cls(self._compute())
+
+    def _compute(self) -> Sequence[Polygon]:
         if not (self.left and self.right):
             return self.left or self.right
-        elif bounding_box.disjoint_with(
-                bounding_box.from_multipolygon(self.left),
-                bounding_box.from_multipolygon(self.right)):
-            result = self.left + self.right
-            result.sort(key=to_first_boundary_vertex)
+        elif bounding.disjoint_with(
+                bounding.from_polygons(self.left,
+                                       context=self.context),
+                bounding.from_polygons(self.right,
+                                       context=self.context)):
+            result = []
+            result += self.left
+            result += self.right
+            result.sort(key=to_first_border_vertex)
             return result
         self.normalize_operands()
-        return events_to_multipolygon(self.sweep())
+        return self.events_to_polygons(self.sweep())
 
     def in_result(self, event: Event) -> bool:
         return (event.outside
                 or not event.from_left and event.is_common_region_boundary)
 
 
-def events_to_multipolygon(events: List[Event]) -> Multipolygon:
-    events = sorted([event for event in events if event.primary.in_result],
-                    key=EventsQueueKey)
-    for index, event in enumerate(events):
-        event.position = index
-    are_internal, depths, holes, parents = [], [], [], []
-    processed = [False] * len(events)
-    contours = []
-    connectivity = events_to_connectivity(events)
-    for index, event in enumerate(events):
-        if processed[index]:
-            continue
-        contour_id = len(contours)
-        _compute_relations(event, contour_id, are_internal, depths, holes,
-                           parents)
-        contour = _events_to_contour(event, events, contour_id, connectivity,
-                                     processed)
-        shrink_collinear_vertices(contour)
-        if depths[contour_id] % 2:
-            # holes will be in clockwise order
-            contour.reverse()
-        contours.append(contour)
-    result = []
-    for index, contour in enumerate(contours):
-        if are_internal[index]:
-            # hole of a hole is an external polygon
-            result.extend((contours[hole_index],
-                           [contours[hole_hole_index]
-                            for hole_hole_index in holes[hole_index]])
-                          for hole_index in holes[index])
-        else:
-            result.append((contour, [contours[hole_index]
-                                     for hole_index in holes[index]]))
-    return result
-
-
 def _compute_relations(event: Event,
                        contour_id: int,
                        are_internal: List[bool],
@@ -373,11 +413,11 @@ def _compute_relations(event: Event,
     are_internal.append(is_internal)
 
 
-def _events_to_contour(cursor: Event,
-                       events: List[Event],
-                       contour_id: int,
-                       connectivity: List[int],
-                       processed: List[bool]) -> Contour:
+def _events_to_contour_vertices(cursor: Event,
+                                events: Sequence[Event],
+                                contour_id: int,
+                                connectivity: Sequence[int],
+                                processed: List[bool]) -> List[Point]:
     contour_start = cursor.start
     contour = [contour_start]
     contour_events = [cursor]
@@ -412,8 +452,8 @@ def _events_to_contour(cursor: Event,
 
 
 def _to_next_position(position: int,
-                      processed: List[bool],
-                      connectivity: List[int]) -> Optional[int]:
+                      processed: Sequence[bool],
+                      connectivity: Sequence[int]) -> Optional[int]:
     candidate = position
     while True:
         candidate = connectivity[candidate]
